@@ -301,35 +301,40 @@ check_start_stream (TpStreamEngineStream *stream)
       if (farsight_stream_get_state (priv->fs_stream) == FARSIGHT_STREAM_STATE_CONNECTED)
         {
           farsight_stream_start (priv->fs_stream);
-
-          if (stream->media_type == FARSIGHT_MEDIA_TYPE_VIDEO)
-            {
-              TpStreamEngine *engine = tp_stream_engine_get ();
-              GstElement *pipeline = tp_stream_engine_get_pipeline (engine);
-              gst_element_set_state (pipeline, GST_STATE_PLAYING);
-            }
-
           priv->stream_started = TRUE;
         }
      }
 }
 
+typedef struct _method_call_ctx method_call_ctx;
+
+struct _method_call_ctx
+{
+  TpStreamEngineStream *stream;
+  const gchar *method;
+};
+
 /* dummy callback handler for async calling calls with no return values */
 static void
-dummy_callback (DBusGProxy *proxy, GError *error, gpointer user_data)
+async_method_callback (DBusGProxy *proxy, GError *error, gpointer user_data)
 {
+  method_call_ctx *ctx = (method_call_ctx *) user_data;
+
   if (error)
     {
-      g_warning ("Error calling %s: %s", (gchar *) user_data, error->message);
+      g_warning ("Error calling %s: %s", ctx->method, error->message);
+      g_signal_emit (ctx->stream, signals[STREAM_ERROR], 0);
       g_error_free (error);
     }
+
+  g_slice_free (method_call_ctx, ctx);
 }
 
 static void
-state_changed (FarsightStream *stream,
-               FarsightStreamState state,
-               FarsightStreamDirection dir,
-               gpointer user_data)
+cb_fs_state_changed (FarsightStream *stream,
+                     FarsightStreamState state,
+                     FarsightStreamDirection dir,
+                     gpointer user_data)
 {
   TpStreamEngineStream *self = TP_STREAM_ENGINE_STREAM (user_data);
   TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
@@ -350,21 +355,26 @@ state_changed (FarsightStream *stream,
 
   if (priv->stream_handler_proxy)
     {
+      method_call_ctx *ctx = g_slice_new0 (method_call_ctx);
+
+      ctx->stream = self;
+      ctx->method = "Media.StreamHandler::StreamState";
+
       tp_media_stream_handler_stream_state_async (
-        priv->stream_handler_proxy, state, dummy_callback,
-        "Media.StreamHandler::StreamState");
+        priv->stream_handler_proxy, state, async_method_callback, ctx);
     }
 }
 
 static void
-new_native_candidate (FarsightStream *stream,
-                      gchar *candidate_id,
-                      gpointer user_data)
+cb_fs_new_native_candidate (FarsightStream *stream,
+                            gchar *candidate_id,
+                            gpointer user_data)
 {
   TpStreamEngineStream *self = TP_STREAM_ENGINE_STREAM (user_data);
   TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
   const GList *fs_candidates, *lp;
   GPtrArray *transports;
+  method_call_ctx *ctx;
 
   fs_candidates = farsight_stream_get_native_candidate (stream, candidate_id);
   transports = g_ptr_array_new ();
@@ -427,9 +437,13 @@ new_native_candidate (FarsightStream *stream,
       g_ptr_array_add (transports, g_value_get_boxed (&transport));
     }
 
+  ctx = g_slice_new0 (method_call_ctx);
+  ctx->stream = self;
+  ctx->method = "Media.StreamHandler::NativeCandidatesPrepared";
+
   tp_media_stream_handler_new_native_candidate_async (
-      priv->stream_handler_proxy, candidate_id, transports, dummy_callback,
-      "Media.StreamHandler::NativeCandidatesPrepared");
+      priv->stream_handler_proxy, candidate_id, transports,
+      async_method_callback, ctx);
 }
 
 /**
@@ -446,7 +460,7 @@ tp_transports_to_fs (gchar* candidate, GPtrArray *transports)
   GList *fs_trans_list = NULL;
   GValueArray *transport;
   FarsightTransportInfo *fs_transport;
-  int i;
+  guint i;
 
   for (i=0; i< transports->len; i++)
     {
@@ -620,7 +634,7 @@ set_remote_candidate_list (DBusGProxy *proxy, GPtrArray *candidates,
   GValueArray *candidate = NULL;
   GPtrArray *transports = NULL;
   gchar *candidate_id = NULL;
-  int i;
+  guint i;
 
   for (i = 0; i < candidates->len; i++)
     {
@@ -663,8 +677,9 @@ set_remote_codecs (DBusGProxy *proxy, GPtrArray *codecs, gpointer user_data)
   GHashTable *params = NULL;
   FarsightCodec *fs_codec;
   GList *fs_params = NULL;
-  int i;
+  guint i;
   GPtrArray *supp_codecs;
+  method_call_ctx *ctx;
 
   DEBUG (self, "called");
 
@@ -711,9 +726,12 @@ set_remote_codecs (DBusGProxy *proxy, GPtrArray *codecs, gpointer user_data)
   supp_codecs = fs_codecs_to_tp (
       farsight_stream_get_codec_intersection (priv->fs_stream));
 
+  ctx = g_slice_new0 (method_call_ctx);
+  ctx->stream = self;
+  ctx->method = "Media.StreamHandler::SupportedCodecs";
+
   tp_media_stream_handler_supported_codecs_async
-    (priv->stream_handler_proxy, supp_codecs, dummy_callback,
-     "Media.StreamHandler::SupportedCodecs");
+    (priv->stream_handler_proxy, supp_codecs, async_method_callback, ctx);
 
   for (lp = g_list_first (fs_codecs); lp; lp = g_list_next (lp))
     {
@@ -744,8 +762,8 @@ stop_stream (TpStreamEngineStream *self)
   if (self->media_type == FARSIGHT_MEDIA_TYPE_VIDEO)
     {
       GstElement *sink = farsight_stream_get_sink (priv->fs_stream);
-      farsight_stream_stop (priv->fs_stream);
       _remove_video_sink (self, sink);
+      farsight_stream_stop (priv->fs_stream);
     }
   else
     {
@@ -829,21 +847,31 @@ prepare_transports (TpStreamEngineStream *self)
   if (priv->got_connection_properties && priv->candidate_preparation_required)
     {
       farsight_stream_prepare_transports (priv->fs_stream);
+      method_call_ctx *ctx;
+
 
       codecs = fs_codecs_to_tp (
                  farsight_stream_get_local_codecs (priv->fs_stream));
 
       DEBUG (self, "calling MediaStreamHandler::Ready");
+
+      ctx = g_slice_new0 (method_call_ctx);
+      ctx->stream = self;
+      ctx->method = "Media.StreamHandler::Ready";
+
       tp_media_stream_handler_ready_async (
-        priv->stream_handler_proxy, codecs, dummy_callback, self);
+        priv->stream_handler_proxy, codecs, async_method_callback, ctx);
     }
 }
 
 static void
-codec_changed (FarsightStream *stream, gint codec_id, gpointer user_data)
+cb_fs_codec_changed (FarsightStream *stream,
+                     gint codec_id,
+                     gpointer user_data)
 {
   TpStreamEngineStream *self = TP_STREAM_ENGINE_STREAM (user_data);
   TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
+  method_call_ctx *ctx;
 
   if (self->media_type == FARSIGHT_MEDIA_TYPE_AUDIO)
     {
@@ -857,9 +885,13 @@ codec_changed (FarsightStream *stream, gint codec_id, gpointer user_data)
       self->stream_id, TRUE);
 
   DEBUG (self, "codec_id=%d, stream=%p", codec_id, stream);
+
+  ctx = g_slice_new0 (method_call_ctx);
+  ctx->stream = self;
+  ctx->method = "Media.StreamHandler::CodecChoice";
+
   tp_media_stream_handler_codec_choice_async (
-    priv->stream_handler_proxy, codec_id, dummy_callback,
-    "Media.StreamHandler::CodecChoice");
+    priv->stream_handler_proxy, codec_id, async_method_callback, ctx);
 }
 
 void
@@ -872,45 +904,53 @@ tp_stream_engine_stream_error (
 
   g_message ("%s: stream errorno=%d error=%s", G_STRFUNC, error, message);
 
-  tp_media_stream_handler_error_async (priv->stream_handler_proxy, error,
-      message, dummy_callback, NULL);
-}
-
-static void
-stream_error (
-  FarsightStream *stream,
-  FarsightStreamError error,
-  const gchar *debug,
-  gpointer user_data)
-{
-  TpStreamEngineStream *self = TP_STREAM_ENGINE_STREAM (user_data);
-  TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
-
-  g_message ("%s: stream error: stream=%p error=%s", G_STRFUNC, stream, debug);
-  /* FIXME: check if error is EOS */
-  tp_media_stream_handler_error (priv->stream_handler_proxy, 0, debug, NULL);
+  tp_media_stream_handler_error (priv->stream_handler_proxy, error,
+      message, NULL);
   g_signal_emit (self, signals[STREAM_ERROR], 0);
 }
 
 static void
-new_active_candidate_pair (FarsightStream *stream, const gchar* native_candidate, const gchar *remote_candidate, gpointer user_data)
+cb_fs_stream_error (FarsightStream *stream,
+                    FarsightStreamError error,
+                    const gchar *debug,
+                    gpointer user_data)
 {
   TpStreamEngineStream *self = TP_STREAM_ENGINE_STREAM (user_data);
-  TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
 
-  DEBUG (self, "stream=%p", stream);
-
-  tp_media_stream_handler_new_active_candidate_pair_async
-    (priv->stream_handler_proxy, native_candidate, remote_candidate, dummy_callback, "Media.StreamHandler::NewActiveCandidatePair");
+  /* FIXME: map Farsight errors to Telepathy errors */
+  tp_stream_engine_stream_error (self, 0, debug);
 }
 
 static void
-native_candidates_prepared (FarsightStream *stream, gpointer user_data)
+cb_fs_new_active_candidate_pair (FarsightStream *stream,
+                                 const gchar* native_candidate,
+                                 const gchar *remote_candidate,
+                                 gpointer user_data)
+{
+  TpStreamEngineStream *self = TP_STREAM_ENGINE_STREAM (user_data);
+  TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
+  method_call_ctx *ctx;
+
+  DEBUG (self, "stream=%p", stream);
+
+  ctx = g_slice_new0 (method_call_ctx);
+  ctx->stream = self;
+  ctx->method = "Media.StreamHandler::NewActiveCandidatePair";
+
+  tp_media_stream_handler_new_active_candidate_pair_async (
+    priv->stream_handler_proxy, native_candidate, remote_candidate,
+    async_method_callback, ctx);
+}
+
+static void
+cb_fs_native_candidates_prepared (FarsightStream *stream,
+                                  gpointer user_data)
 {
   TpStreamEngineStream *self = TP_STREAM_ENGINE_STREAM (user_data);
   TpStreamEngineStreamPrivate *priv = STREAM_PRIVATE (self);
   const GList *transport_candidates, *lp;
   FarsightTransportInfo *info;
+  method_call_ctx *ctx;
 
   DEBUG (self, "stream=%p", stream);
 
@@ -923,9 +963,13 @@ native_candidates_prepared (FarsightStream *stream, gpointer user_data)
         (info->proto == FARSIGHT_NETWORK_PROTOCOL_TCP) ? "TCP" : "UDP",
         info->proto_subtype, info->ip, info->port, (double) info->preference);
   }
+
+  ctx = g_slice_new0 (method_call_ctx);
+  ctx->stream = self;
+  ctx->method = "Media.StreamHandler::NativeCandidatesPrepared";
+
   tp_media_stream_handler_native_candidates_prepared_async (
-    priv->stream_handler_proxy, dummy_callback,
-    "Media.StreamHandler::NativeCandidatesPrepared");
+    priv->stream_handler_proxy, async_method_callback, ctx);
 }
 
 static void
@@ -1144,7 +1188,8 @@ tp_stream_engine_stream_go (
   gchar *conn_timeout_str;
 
 #ifdef MAEMO_OSSO_SUPPORT
-  media_server_proxy_init (stream);
+  if (media_type == TP_MEDIA_STREAM_TYPE_AUDIO)
+    media_server_proxy_init (stream);
 #endif
 
   priv->channel_path = g_strdup (channel_path);
@@ -1210,18 +1255,18 @@ tp_stream_engine_stream_go (
     }
 
   g_signal_connect (G_OBJECT (priv->fs_stream), "error",
-                    G_CALLBACK (stream_error), stream);
+                    G_CALLBACK (cb_fs_stream_error), stream);
   g_signal_connect (G_OBJECT (priv->fs_stream), "new-active-candidate-pair",
-                    G_CALLBACK (new_active_candidate_pair), stream);
+                    G_CALLBACK (cb_fs_new_active_candidate_pair), stream);
   g_signal_connect (G_OBJECT (priv->fs_stream), "codec-changed",
-                    G_CALLBACK (codec_changed), stream);
+                    G_CALLBACK (cb_fs_codec_changed), stream);
   g_signal_connect (G_OBJECT (priv->fs_stream), "native-candidates-prepared",
-                    G_CALLBACK (native_candidates_prepared), stream);
+                    G_CALLBACK (cb_fs_native_candidates_prepared), stream);
   priv->state_changed_handler_id =
     g_signal_connect (G_OBJECT (priv->fs_stream), "state-changed",
-                      G_CALLBACK (state_changed), stream);
+                      G_CALLBACK (cb_fs_state_changed), stream);
   g_signal_connect (G_OBJECT (priv->fs_stream), "new-native-candidate",
-                    G_CALLBACK (new_native_candidate), stream);
+                    G_CALLBACK (cb_fs_new_native_candidate), stream);
 
   /* OMG, Can we make dbus-binding-tool do this stuff for us?? */
 
